/************************************************************ 
SND2SPU.C v 2.1	(author  Rob Vawter[email: devtech_support@playstation.sony.com])
copyright 1997 Sony Computer Entertainment America
	ALL RIGHTS RESERVED

History: 
	2.1	11/18/97 Changed FindOpenVoice() to work with one-shot VAGs 
	2.0 11/13/97 Changed some #define names and included new #ifdef statements
				 in SFX_data_struct for compilation purposes
	1.3 11/11/97 Fixed 4 errors 
		SpuSetKey() param and SpuSetReverbVoice() param orders were reversed

	1.2 Added new level of usage and replacement comments 

	THIS IS UNPUBLISHED PROPREITRARY SOURCE CODE OF SONY COMPUTER 
	ENTERTAINMENT AMERICA
	This code is sample code. Sony Computer Entertainment America and 
	the author(s) accept no responsibility for any problems which may 
	arise from using this code. This file may be used and/or modified 
	provided the disclaimer mentioned above is maintained, and the 
	appropriate Non Disclosure Agreement with Sony Computer Entertainment 
	America and the party or parties using/modifying the code exist. 
	This file may not be distributed to any party/parties which do not
	have the appropriate Non Disclosure Agreement with Sony Computer Entertainment America
**************************************************************/
/* Uses:
	Code contained within this documented can be used to help substitute 
	libspu calls for previously used libsnd calls in cases where libsnd
	is being used with no MIDI interpretation */

/* I realize that the structure of the sound calls here will not exactly match
   every program - you may still keep your structure and substitute the specified calls	
   (grep on substitute)
   */


/* Why change to spu if not using MIDI? 

An overview
-----------
	The SPU registers are very, very slow (2-3 times slower to access than main RAM due to SPU bus speed)
Accessing these registers unnecessarily is BAD. So, the code in here is trying to be designed to do
this as little as possible and to use the libspu functions which have the greatest speed and/or the 
smallest size (for i-cache purposes)
	Libsnd does a wonderful job of playing back MIDI; however it is a bulky library, with overkill
code in some places. 

	1) Pitch calculation is very slow and very bulky. The SPU registers expect a pitch value, 
	but libsnd must convert from a sample note and key on note (with appropriate fine tuning) 
	which uses a lookup table (770 bytes big), but is math intensive first. Libspu allows for 
	direct pitch input. Pitch could also not be easily set to 0 in pre 4.1 libsnd. 
	New 4.1 functions can workaround these problems, but there are many other issues.

	2) Volume calculation for libsnd keyon takes two input volumes (left and right) from 0-127.
	These are converted into range 0-0x3fff via 8 mults and 6 divides - *ow*. 
	New libsnd 4.1 functions allow for greater volume range (-0x4000-0x3fff, for 3d sound) 
	and user volume calcs. But libspu gives even greater freedom, such as negative volumes 
	for reverb depth (allowing for 3d echo location) and linear and exponential increase and 
	decrease modes, which change a volume from init val to 0x3fff or 0.
  
	3) The tick callback is a killer. It's wonderful if it is used for interpreting MIDI commands,
	but the probable i-cache miss isn't worth it for sound effects alone. Libspu will avoid this.	  
	
	4) Main RAM. Libspu has less global variables, less code size, less header info for 
	VAG playback (which is also stored moe compactly for a double savings), and less 
	Spu memory management size (depending on how you do things)

	5) Design your own faster or just plain different voice allocation system - 
	multiple voice, consecutive voices, whatever.

	6) You'll even take less loading time - woohoo! 
	(I know that's the real reason you're reading this doc)
	(Well okay maybe not, but it's a nice side effect)

*/

/* How to change key on calls - an overview:

1)  The key on calls in libsnd must be replaced by several layers of code
a) Voice Allocation Function
	This function should be based on empty voices and any other parameters necessary. Libsnd
	allocates the lowest priority, smallest envelope (in case of ties on priority) , 
	and oldest keyed on sound (in case of envelope ties), when all voices are used.
b) Volume Calculation  
  your volumes can be: 
  i) all location based (3d or left-right pan) 
  ii) all fixed or ambient sounds 
  iii) a mixture of the above two
  This said, your volume calculations should be as simple as possible, to avoid wasted cycles,
  and should only be performed when necessary (ex: shotgun sound in 1 player DOOM should never
  be calculated at keyon - its always the same volume and same pan)
c) Voice Settings
	These include volume left and right, adsr1 and adsr2, playback rate (pitch or sample note and 
	key on note) and SPU RAM address of VAG start. Additionally, reverb may be applied.
d) Actual key on


/* BEGIN REPLACEMENT TYPE CODE HERE */

/* Global type stuff */

/*To replace VAB header you would need to store the following info:
  it's faster to preset the pitch of each sfx than calc it every time.
  ditto with volume (although you may apply relative left/right distance effects to 
  non-ambient sounds)	 

  I'm still working on a utility which will convert existing sfx VAB headers to a new
  libspu-friendly (and much smaller) data format  */
#include <libspu.h>
// #define QUEUEING 
					/* if using MIDI engine or very,very active sfx engine in libspu, queueing is
                     a good idea to ensure that keyons/offs which occur more frequently than
					 speed of SPU (1/44100 sec) are not missed - if criteria above not met,
					 queueing is cumbersome and annoying  */ 

#define PAN_USED /* see note in vh2spu.c utility source code */
#define MODE_USED /* will reverb be used and will it be applied to only SOME voices? If yes, 
						you need this. Otherwise you don't! */
#ifndef MODE_USED
// remove the following line if reverb is always off in your program
#define REVERB_ALWAYS_ON 

#endif
#define PRIORITY_USED
					/* priority of voices used by your code - remove if your maximum number
					   of sfxs which can be playing simultaneously if less than the number 
					   of voices you will allocate for sfx; priority is only needed for 
					   voice overflow */
#define SFX_LOADED_ONLY_ONCE 
					/* Are all sfx loaded at the beginning of the level? If multiple
					loads occur, remove this line */
								


#define NUM_SFX_VAGS 2 	// Check VAB header - number of vags
#define NUM_SFX 3 		/* Number of tones in VAB header - differs from above since
				    		multiple sfx may use the same vag */
#define NUM_SFX_VOICES 24 /* Number of voices you wish to allocate for sfx
					       	Ex: if using one VAG streaming voice for music, 
					       	NUM_SFX_VOICES = 23 
					       	If using XA or DA music 
					       	NUM_SFX_VOICES = 24*/		

unsigned long vag_offset_table[NUM_SFX_VAGS] = {0,0x48D80};
		// <<3 the vag offset table in the vab header
		// or fill from the srcvabfilename.off file which the utility vh2spu creates

/* Ensure if any members of structure removed that proper size alignment is followed */
/* Padding these out to longs wasn't much fun, 
please feel free to do it differently */
typedef struct SFX_data_struct{
#ifdef PRIORITY_USED
unsigned char 	prior;
#else	//PRIORITY_USED
#define PRI_NOT_USED
#endif  //PRIORITY_USED

#ifdef MODE_USED
unsigned char 	mode;
#ifdef PRI_NOT_USED
#define PAD_ONE	  // only 1 of mode and pri used - 1 pad on this long
#else  // PRI_NOT_USED
#define PAD_ZERO  // both mode and pri used - no pad on this long
#endif // PRI_NOT_USED
#else //if not MODE_USED
#ifdef PRI_NOT_USED
#define PAD_CARRYOVER  // neither mode nor pri used - carry padding to PAN
#else //PRI_NOT_USED
#define PAD_ONE	  // only 1 of mode and pri used - 1 pad on this long
#endif //PRI_NOT_USED
#endif //MODE_USED


#ifdef PAD_ONE
unsigned char	pad1;
#endif //PAD_ONE

short		  	vol;


#ifdef PAN_USED
unsigned char 	pan;
#if defined(PAD_ZERO) || defined(PAD_ONE)
unsigned char	pad2;
unsigned short	pad3;
#endif // PAD_ZERO || PAD_ONE
#ifdef PAD_CARRYOVER
unsigned char	pad4;
#endif //PAD_CARRY
#else // not PAN_USED
#ifdef PAD_CARRYOVER
unsigned short	pad5;
#endif //PAD_CARRY
#endif //PAN_USED

unsigned short	pitch;
unsigned short	adsr1;

unsigned short	adsr2;
short			vagnum;

} SFX_data_struct;





#ifdef NOVEMEBER_11_VER_ONE_PT_THREE 
/* this section of removed code contains info about how to build your own util
 and did not include proper long word alignment of structure*/

typedef struct SFX_data_struct {
#ifdef PRIORITY 
char priority; // fill straight from vab header - not needed if no voice priority used in your voice allocation
#endif
#ifdef REVERB_CHECK
char mode; // fill straight from vab header - 0 == normal, 4 == reverb
#endif
short vol; /* take the tone and program volumes as percentages of 127 and reduce 
				the multiply * 0x3fff (max) 
			   Ex: tone volume == 64, program volume == 127. This sound
			   was intended to be half as loud as other tones in the program. This 
			   program was intended to be at max volume as compared to other
			   programs. You may also factor in master VAB volume if you have more than 
			   one sfx VAB which would be loaded in and the two master VAB volumes 
			   differ (two different master VAB volumes means that all sounds from one VAB are
			   intended to be lower in volume than the other)
			   Ex result: main volume = 0x3fff */		       
short pan	 /* You may factor pan in if you wish, but figuring volume calculations
                via object location will set up pan values and scale volumes down for 
                things like distance away. Only used for ambient sounds or sounds which always
				have the same location (player shotgun in Doom for example) */
short pitch; /* My util handles this but you may need info from the musician to properly
				convert */

short adsr1; /* fill straight from vab header */
short adsr2; /* fill straight from vab header */
int vag_number; /* Which vag is the sound effect playing back? - 
				   in vab header they number from 1-254
				   Take this into account - either renumber 0-253 (easiest for me) 
				   OR always remember to subtract one 
				   when using array vag_addr_in_SPUram below and vag_offset_table array above
				   I assume renumbering in my util*/
} SFX_data_struct;
#endif // NOVEMEBER_11_VER_ONE_PT_THREE 

SFX_data_struct	SFX_data[NUM_SFX];


unsigned long vag_addr_in_SPUram[NUM_SFX_VAGS]; 
	/* gets set by using return value of first vag and then 
	 ...comparing that to vag offset table */

#ifdef PRIORITY
char voice_priority[NUM_SFX_VOICES]; 
// may not be used if you don't have priority in voice allocation function


int FindOpenVoice(char); 
#else /* PRI */
int FindOpenVoice(void); 
#endif /* PRI */
void InitSound(void); 
int TransferSoundsToSPURAM(unsigned char*, unsigned long);
int KeyOnSFX(int, long);
void YourVolCalc(long, short*, short*, int);
void UpdateSFXVol(int, int, long);
void InitReverbOn(long, short, short);


#ifndef SFX_LOADED_ONLY_ONCE 
/* No SpuMalloc() calls needed if you load all sfx in at beginning of level and load them
   all as one file and know whether whatever reverb effect you may want to use will also fit */

#define SPU_MALLOC_MAX 	1 
//if sound only loaded once per level - can load all VAGs as one file

char spu_malloc_rec [SPU_MALLOC_RECSIZ * (SPU_MALLOC_MAX + 1)]; 
// libsnd reserves 16 or 17 loads - spu malloc table smaller on main RAM this way

#endif /* SFX_LOADED_ONLY_ONCE */

/* this must be called immediately after CdInit() (why? to remove the reverb echo
   caused by Sony boot-up logos; otherwise reverb echo will remain thru entire
   movies like WipeOutXL or Test Drive 4) and before transfer of waveform data */
void InitSound() 
{
#ifdef QUEUEING 
/* See note below on queueing */
 SpuEnv spuqueue; 
#endif

	/* init sound lib */
	/* substitute for SsInit() */
	SpuInit();

#ifdef SFX_LOADED_ONLY_ONCE 
	/* substitutes for nothing directly in libsnd - SsInit() normally calls this - you must 
	if statements in DEFINES area met */
	SpuInitMalloc (SPU_MALLOC_MAX, spu_malloc_rec);
#endif /* SFX_LOADED_ONLY_ONCE */

	/* Set master left and right volumes of spu to maximum (range is -0x4000 ~ 0x3fff),
	   as opposed to libsnd (0-127) */
	/* substitute for SsSetMVol(127,127) */
	SpuSetCommonMasterVolume(0x3fff, 0x3fff); 

	/* Turn cd mixing on */
	/* substitue for SsSetSerialAttr(SS_SERIAL_A, SS_MIX, SS_SON) */
	SpuSetCommonCDMix(SPU_ON); 

	/* Set l & r cd volumes to midrange (range is -0x8000 ~ 0x7fff) */
	/* substitue for SsSetSerialVol(SS_SERIAL_A, 127, 127) */
	SpuSetCommonCDVolume(0x3fff, 0x3fff);
	/* Note: use CdMix() to set CD input to mono or stereo */

	/* Spu queueing - keyon/off, pitchmodulation, noise generator, 
	   and reverb on/off calls will all be queued up UNTIL SpuFlush() is called. 

	   Queuing accomplishes two things 
	   1): makes sure that keyon/offs which happen very very quickly will not be missed 
	   (if too many happen in 1/44000 second they may be missed - SPU hardware speed issue 
	   - really only a concern for libspu MIDI engines)

	   2) Allows for reallocation of a voice without ever turning on the voice - so, no noise clips
	    and fewer register settings.

	   Queueing caveats: 
	   1) Must be careful changing volumes, pitch, adsr of sounds which have not yet been keyed on 
	   and are awaiting flush - otherwise "old" waveform may the one getting the changes. 
	   
	   workaround#1: do a voice check to make sure the vag being affected is the proper vag. i.e.
	   pass the sfx_num of the vag being affected and store the sfx_num of the voice in a global
	   array.
	   
	   2) Don't forget to call SpuFlush() in your main loop! (once per frame unless running
	    at very slow rate should do - don't "overupdate" sound - visuals are far more noticeable
	    than sounds) */
	   
#ifdef QUEUEING 
	   /* Queueing init */
	    spuqueue.mask = SPU_ENV_EVENT_QUEUEING;
	    spuqueue.queueing = SPU_ON;
	    SpuSetEnv(spuqueue);
#endif	   
}

/* TransferSoundsToSPURAM - this function does work which can substitute for 
   SsVabOpenHead() ->  SsVabTransBody() -> SsVabTransCompleted() series [also SsVabTransfer, 
   which is not recommended because it is a blocking function - use the transfer time to 
   perform other work!] */
/* parameters : vag_start_addr_main_ram - where do the VAGs live in main RAM
				vag_body_size - how big are the VAGs */
int TransferSoundsToSPURAM(unsigned char* vag_start_addr_main_ram, unsigned long vag_body_size)
{
long vag_start_addr_in_spu_ram;
unsigned long tsa;
unsigned long spu_written_size;
int i;
int ret = 0;

#ifndef SFX_LOADED_ONLY_ONCE 
	/* Allocate area in SPU RAM for VAGs */
 	if (vag_start_addr_in_spu_ram = SpuMalloc(vag_body_size) == -1)
	{
	 printf("spu malloc failed");
	 return(-2);
	}
#else vag_start_addr_in_spu_ram = 0x1010;
#endif	
	/* Set start address for transfer to SPU RAM */
	/* libsnd performs both this and SpuMalloc() in SsVabOpenHead() - so these 2 + see below
	 substitute for SsVabOpenHead() */
	tsa = SpuSetTransferStartAddr(vag_start_addr_in_spu_ram);

	/* Probably unneeded call - transfer mode default IS DMA - this call only 
	   needed if transfer mode is explicitly set to I/O */
	// SpuSetTransferMode(SPU_TRANSFER_BY_DMA);

	/* Transfer waveform data - substitutes for SsVabTransBody() 
	May also be done in parts by SpuWritePartly() to substitute 
	for SsVabTransPartly()*/
	if (spu_written_size = SpuWrite(vag_start_addr_main_ram, vag_body_size) != vag_body_size) 
	{
		ret = -1;
		printf("fullsize not transferred");
	}

	/* fill array keeping track of individual vag addresses in SPU RAM 
	libsnd sets this up internally in VAB header */
	for (i=0; i++; i< NUM_SFX_VAGS)
	{
	 vag_addr_in_SPUram[i] = vag_start_addr_in_spu_ram;
	 vag_start_addr_in_spu_ram += vag_offset_table[i];
	}

	/* could also insert other functions here - otherwise the time of SPU RAM load time is wasted 
	(up to 1 second for all 512 K);  longer loads == less happy consumers */

	/* Check that transfer is finished - could also have set up callback before transfer start */
	SpuIsTransferCompleted (SPU_TRANSFER_WAIT);

	return(ret);
}

/* Keying on SFX */
/* this whole functions substitutes for SsUtKeyOn() or SsUtKeyOnV()+ your voice allocator */ 
/* SsVoKeyOn may turn on more than one sfx tone at a time - 
    you can set this func. up to do that as well */
int KeyOnSFX(int sfx_num, long your_location_params)
{
int voice_num;	
short voll, volr;

	/* Allocate Voice(s) - done internally by SsUtKeyOn() */
	if (voice_num = FindOpenVoice(SFX_data[sfx_num].priority) == -1)	
	{
		return(-1);
	}
	else
	{
	 /* Your vol calc only necessary for sounds which need to be located in space - 
	 left-right pan or distance effects - so things like laserblast of player, ambient sounds
	 need NO pre-vol calc */
	 YourVolCalc(your_location_params, &voll, &volr, sfx_num);

	 /* set volumes -0x4000 ~ 0x3fff */ 
	 SpuSetVoiceVolume(voice_num, voll, volr);

	 /*	If ALL of your sounds will use the same pitch and no pitch modulation will be used, 
		init all sfx allocatable voices to that pitch at beginning of level and 
		never call this function again (yes, I realize that this is a bit limiting, 
		but hey, some people do it)*/
	 SpuSetVoicePitch(voice_num, SFX_data[sfx_num].pitch);

	 /* set waveform start address in SPU RAM - this is why you set up that offset in
		TansferSoundstoSPURAM */
	 SpuSetVoiceStartAddr(voice_num,vag_addr_in_SPUram[sfx_num]);

	 /* see VAB header format post on web for more adsr1 and adsr2 conversion 
	 - no voodoo here, really! - you MAY want to check MY bit math however */
	 /* abbreviated adsr info
 		adsr1 bits	 
		15 ar_m: 	attack rate mode 	0=linear, 1=exponential
		8-14ar: 	attack rate 
		4-7dr: 		decay rate
		0-3sl: 		sustain level

		adsr2 bits
		15sr_m: 	sustain rate mode	0=linear, 1=exponential
		14sr_s: 	sustain rate sign	0=positive, 1=negative
		13: 		unused
		6-12sr:		sustain rate
		5rr_m:		release rate mode	0=linear, 1=exponential
		0-4rr:		release rate */
	 SpuSetVoiceADSR(voice_num,(SFX_data[sfx_num].adsr1>>8) & 0x7f, (SFX_data[sfx_num].adsr1>>4) & 0x8,
	 (SFX_data[sfx_num].adsr2>>6) & 0x7f, (SFX_data[sfx_num].adsr2>>5) & 1, (SFX_data[sfx_num].adsr2) & 0xf);
	 /* or 	 SpuSetVoiceADSRAttr(voice_num,)
	 which allows mode settings - default modes are 0 for all modes */

	 /* in 4.2 substitute the next two calls for the above (mucho faster and easier) */
	 //
	 //
	 // SpuSetVoiceADSR1(voice_num, SFX_data[sfx_num].adsr1);
	 // SpuSetVoiceADSR2(voice_num,	SFX_data[sfx_num].adsr2);

	 /* Set reverb on or off as applicable; this only includes sounds which ALWAYS have reverb - 
	    environmental reverb not covered (ie tunnel makes ALL sfx have reverb) - add code for this */
#ifdef REVERB_CHECK
	 if (SFX_data[sfx_num].mode & 0x4) SpuSetReverbVoice(SPU_ON, 1<<voice_num);
	 else SpuSetReverbVoice(SPU_OFF, 1<<voice_num);
#endif 
	 /* Key voice on */
	 SpuSetKey(SPU_ON,1<<voice_num);
	 return(0);
	}
}


/* Function - allocates an open voice   
(just a sample - this could be simplified or include a way to 
have consecutive voice nums - whatever YOU want) */
#ifdef PRIORITY
int FindOpenVoice(char new_priority) 
#else
int FindOpenVoice()
#endif
{
#ifdef PRIORITY
int lowpri_voice = NUM_SFX_VOICES+1;
int lowest_priority = new_priority;
#endif
int empty_voice = NUM_SFX_VOICES+1;
int i;

/* if you are using SpuSetEnv() and SpuFlush() to queue keyon commands, you would be MUCH better
 off cycle-wise to make voice_status global, call SpuGetAllKeysStatus() once per VBlank, 
   and changing the array as YOU fill voices in [I chose #define SPU_VOICE_ASSIGNED 4 as my 
   macro for since it is one greater than SPU_ON_ENV_OFF*/
#ifdef QUEUEING
#define SPU_VOICE_ASSIGNED 4 
#else
char voice_status[NUM_SFX_VOICES]; /* this would be GLOBAL ifdef QUEUEING */
#endif

#ifndef NOTHING_ELSE_WORKED
	/* check status of voices */
	SpuGetAllKeysStatus(voice_status);

	for (i=0; i<NUM_SFX_VOICES; i++)
	{
#ifndef NOV_EIGHTEEN_CHANGE
/* only works if all sounds keyed off by software, not if they expie by hardware (ie one-shot
VAG never return to SPU_OFF) */
		if (voice_status[i] == SPU_OFF) 
#else // NOV_EIGHTEEN_CHANGE
// untested!!!!	- please see scea programming.sound newsgroup

		if ((voice_status[i] == SPU_OFF) || (voice_status[i] == SPU_ON_ENV_OFF))
#endif //NOV_EIGHTEEN_CHANGE
		{
			empty_voice = i;
			break;
		}
#ifdef PRIORITY
		else if (lowest_priority>voice_priority[i])
		{
			lowpri_voice = i;
			lowest_priority = voice_priority[i];
		}													
#endif
	}
	if (empty_voice !=NUM_SFX_VOICES+1)
	{
#ifdef PRIORITY
	  voice_priority[empty_voice] = new_priority;
#endif
#ifdef QUEUEING
	   voice_status[empty_voice] = SPU_VOICE_ASSIGNED;  
#endif
	  return(empty_voice);
	}
#ifdef PRIORITY
	else if (lowpri_voice !=NUM_SFX_VOICES+1)
	{
	  voice_priority[lowpri_voice] = new_priority;
#endif
#ifdef QUEUEING
	   voice_status[lowpri_voice] = SPU_VOICE_ASSIGNED; 
#endif
	  return(lowpri_voice);
	}							   
	return(-1);
#else  NOTHING_ELSE_WORKED
	/* just add one to the last keyed on voice and rol thru the voices one at a time;  
	   when 23 reached, set cur_voice = 0 again and start rolling from the bottom all
	   over again */
#endif
}

/*Updating SFX volumes
 voice_num must be 0-23, not bitfield */
/* this could substitute for many libsnd calls */
void UpdateSFXVol(int voice_num, int sfx_num, long your_location_params)
{
short voll, volr;

YourVolCalc(your_location_params, &voll, &volr, sfx_num); 
/* you may be better off calling this funct for all sfx and storing the results in an array */

/* and then calling this funct for all sfx	for better icache results*/	
SpuSetVoiceVolume(voice_num, voll, volr);  
}

void YourVolCalc(long your_location_params, short *voll, short *volr, int sfx_num)
{
 /* blah blah blah, yeah, whatever
  	you may want locate the sound-emitting object in left-right pan and/or
	reduce volume if the emitter is behind the "ears" or for environmental effects 
	(thin walls, underwater) or may wish for a full 3-d effect.
	This calculation is ENTIRELY up to you - I just use the default values set in the 
	SFX_data_struct member vol (as if these were ambient,unpositioned sounds) 
	I'll put together another doc and sample code with some 3d sounds 
	and reverb effects hard-coded in */
	*voll = SFX_data[sfx_num].vol;
	*volr = SFX_data[sfx_num].vol;
	/* my volume is a set value
	   You may want to have vol be a reduction percentage instead (0x3fff/vol) */
	return;
}
														 
/* reverb on - could also add SpuSetReverbModeDelayTime() and SpuSetReverbModeFeedback() */
void InitReverbOn(long rev_mode, short rvoll, short rvolr)
{
	/* set reverb depth to 0 off to avoid noise - SpuSetReverbDepth() is old and ugly and slow
	use this one instead */
	SpuSetReverbModeDepth(0,0);

	/* turn reverb on for all of SPU - however, this call does NOT 
	   APPLY this reverb to any voices or the cd input
	   Specific function calls must be made to perform reverb application
	   SpuSetReverbVoice() and SpuSetCommonCDReverb() */
	SpuSetReverb(SPU_ON);

	/* Clear reverb work area to help avoid reverb noise 
   	synchronous dma transfer so things like drawing and sound gen
   	will not be performed for a period of time dependent on size of
   	rev_mode */
	SpuSetReverbModeType(rev_mode | SPU_REV_MODE_CLEAR_WA);

#ifndef REVERB_CHECK
/* if reverb always on for all voices or always off for all voices, set it now */
#ifdef REVERB_ALWAYS_ON
 	SpuSetReverbVoice(SPU_ON, SPU_ALLCH);
#else
 	SpuSetReverbVoice(SPU_OFF, SPU_ALLCH);
#endif
#endif

	/* set reverb depth to specified amount */
	SpuSetReverbModeDepth(rvoll, rvolr);
}